from __future__ import annotations

from dataclasses import dataclass
from functools import _lru_cache_wrapper, partial, wraps
from re import findall
from types import (
    BuiltinFunctionType,
    FunctionType,
    MethodDescriptorType,
    MethodType,
    MethodWrapperType,
    WrapperDescriptorType,
)
from typing import TYPE_CHECKING, Any, TypeVar, cast, overload

from typing_extensions import ParamSpec, override

if TYPE_CHECKING:
    from collections.abc import Callable

_P = ParamSpec("_P")
_T = TypeVar("_T")
_U = TypeVar("_U")


def ensure_not_none(obj: _T | None, /, *, desc: str = "Object") -> _T:
    """Ensure an object is not None."""
    if obj is None:
        raise EnsureNotNoneError(desc=desc)
    return obj


@dataclass(kw_only=True, slots=True)
class EnsureNotNoneError(Exception):
    desc: str = "Object"

    @override
    def __str__(self) -> str:
        return f"{self.desc} must not be None"


def first(pair: tuple[_T, Any], /) -> _T:
    """Get the first element in a pair."""
    return pair[0]


@overload
def get_class(obj: type[_T], /) -> type[_T]: ...
@overload
def get_class(obj: _T, /) -> type[_T]: ...
def get_class(obj: _T | type[_T], /) -> type[_T]:
    """Get the class of an object, unless it is already a class."""
    return obj if isinstance(obj, type) else type(obj)


def get_class_name(obj: Any, /) -> str:
    """Get the name of the class of an object, unless it is already a class."""
    return get_class(obj).__name__


def get_func_name(obj: Callable[..., Any], /) -> str:
    """Get the name of a callable."""
    if isinstance(obj, BuiltinFunctionType):
        return obj.__name__
    if isinstance(obj, FunctionType):
        name = obj.__name__
        pattern = r"^.+\.([A-Z]\w+\." + name + ")$"
        try:
            (full_name,) = findall(pattern, obj.__qualname__)
        except ValueError:
            return name
        return full_name
    if isinstance(obj, MethodType):
        return f"{get_class_name(obj.__self__)}.{obj.__name__}"
    if isinstance(
        obj,
        MethodType | MethodDescriptorType | MethodWrapperType | WrapperDescriptorType,
    ):
        return obj.__qualname__
    if isinstance(obj, _lru_cache_wrapper):
        return cast(Any, obj).__name__
    if isinstance(obj, partial):
        return get_func_name(obj.func)
    return get_class_name(obj)


def get_func_qualname(obj: Callable[..., Any], /) -> str:
    """Get the qualified name of a callable."""
    if isinstance(
        obj, BuiltinFunctionType | FunctionType | MethodType | _lru_cache_wrapper
    ):
        return f"{obj.__module__}.{obj.__qualname__}"
    if isinstance(
        obj, MethodDescriptorType | MethodWrapperType | WrapperDescriptorType
    ):
        return f"{obj.__objclass__.__module__}.{obj.__qualname__}"
    if isinstance(obj, partial):
        return get_func_qualname(obj.func)
    return f"{obj.__module__}.{get_class_name(obj)}"


def identity(obj: _T, /) -> _T:
    """Return the object itself."""
    return obj


def is_none(obj: Any, /) -> bool:
    """Check if an object is `None`."""
    return obj is None


def is_not_none(obj: Any, /) -> bool:
    """Check if an object is not `None`."""
    return obj is not None


def not_func(func: Callable[_P, bool], /) -> Callable[_P, bool]:
    """Lift a boolean-valued function to return its conjugation."""

    @wraps(func)
    def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> bool:
        return not func(*args, **kwargs)

    return wrapped


def second(pair: tuple[Any, _U], /) -> _U:
    """Get the second element in a pair."""
    return pair[1]


__all__ = [
    "EnsureNotNoneError",
    "ensure_not_none",
    "first",
    "get_class",
    "get_class_name",
    "get_func_name",
    "get_func_qualname",
    "identity",
    "is_none",
    "is_not_none",
    "not_func",
    "second",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import uuids

from utilities.uuid import UUID_PATTERN

if TYPE_CHECKING:
    from uuid import UUID


class TestUUIDPattern:
    @given(uuid=uuids())
    def test_main(self, *, uuid: UUID) -> None:
        assert UUID_PATTERN.match(str(uuid))



from __future__ import annotations

import datetime as dt
from asyncio import run, sleep
from re import search
from typing import TYPE_CHECKING, Any

from hypothesis import Phase, given, settings
from pytest import mark, param, raises

from tests.conftest import FLAKY
from utilities.asyncio import (
    _MaybeAwaitableMaybeAsyncIterable,
    is_awaitable,
    sleep_dur,
    stream_command,
    timeout_dur,
    to_list,
    try_await,
)
from utilities.datetime import MILLISECOND, duration_to_timedelta
from utilities.hypothesis import durations
from utilities.pytest import skipif_windows
from utilities.timer import Timer

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterable, Iterator

    from utilities.types import Duration

_STRS = list("AAAABBBCCDAABB")


def _get_strs_sync() -> Iterable[str]:
    return iter(_STRS)


async def _get_strs_async() -> Iterable[str]:
    return _get_strs_sync()


def _yield_strs_sync() -> Iterator[str]:
    return iter(_get_strs_sync())


async def _yield_strs_async() -> AsyncIterator[str]:
    for i in _get_strs_sync():
        yield i
        await sleep(0.01)


class TestIsAwaitable:
    @mark.parametrize(
        ("obj", "expected"), [param(sleep(0.01), True), param(None, False)]
    )
    async def test_main(self, *, obj: Any, expected: bool) -> None:
        result = await is_awaitable(obj)
        assert result is expected


class TestSleepDur:
    @FLAKY
    @given(
        duration=durations(
            min_number=0.0,
            max_number=0.01,
            min_timedelta=dt.timedelta(0),
            max_timedelta=10 * MILLISECOND,
        )
    )
    @settings(max_examples=1, phases={Phase.generate})
    async def test_main(self, *, duration: Duration) -> None:
        with Timer() as timer:
            await sleep_dur(duration=duration)
        assert timer >= duration_to_timedelta(duration / 2)

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_dur()
        assert timer <= 0.01


class TestStreamCommand:
    @skipif_windows
    async def test_main(self) -> None:
        output = await stream_command(
            'echo "stdout message" && sleep 0.1 && echo "stderr message" >&2'
        )
        await sleep(0.01)
        assert output.return_code == 0
        assert output.stdout == "stdout message\n"
        assert output.stderr == "stderr message\n"

    @skipif_windows
    async def test_error(self) -> None:
        output = await stream_command("this-is-an-error")
        await sleep(0.01)
        assert output.return_code == 127
        assert output.stdout == ""
        assert search(
            r"^/bin/sh: (1: )?this-is-an-error: (command )?not found$", output.stderr
        )


class TestTimeoutDur:
    @FLAKY
    @given(
        duration=durations(
            min_number=0.0,
            max_number=0.01,
            min_timedelta=dt.timedelta(0),
            max_timedelta=10 * MILLISECOND,
        )
    )
    @settings(max_examples=1, phases={Phase.generate})
    async def test_main(self, *, duration: Duration) -> None:
        with raises(TimeoutError):
            async with timeout_dur(duration=duration):
                await sleep_dur(duration=2 * duration)


class TestToList:
    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_main(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = await to_list(iterable)
        assert result == _STRS


class TestTryAwait:
    async def test_sync(self) -> None:
        def func(*, value: bool) -> bool:
            return not value

        result = await try_await(func(value=True))
        assert result is False

    async def test_async(self) -> None:
        async def func(*, value: bool) -> bool:
            await sleep(0.01)
            return not value

        result = await try_await(func(value=True))
        assert result is False

    @mark.parametrize("cls", [param(ValueError), param(TypeError)], ids=str)
    async def test_error_std_msg_sync(self, *, cls: type[Exception]) -> None:
        def func(*, value: bool) -> bool:
            if not value:
                msg = f"Value must be True; got {value}"
                raise cls(msg)
            return not value

        with raises(cls, match="Value must be True; got False"):
            _ = await try_await(func(value=False))

    @mark.parametrize("cls", [param(ValueError), param(TypeError)], ids=str)
    async def test_error_std_msg_async(self, *, cls: type[Exception]) -> None:
        async def func(*, value: bool) -> bool:
            if not value:
                msg = f"Value must be True; got {value}"
                raise cls(msg)
            await sleep(0.01)
            return not value

        with raises(cls, match="Value must be True; got False"):
            _ = await try_await(func(value=False))

    @mark.parametrize("cls", [param(ValueError), param(TypeError)], ids=str)
    async def test_error_non_std_msg_sync(self, *, cls: type[Exception]) -> None:
        def func(*, value: bool) -> bool:
            if not value:
                raise cls(value)
            return not value

        with raises(cls, match="False"):
            _ = await try_await(func(value=False))

    @mark.parametrize("cls", [param(ValueError), param(TypeError)], ids=str)
    async def test_error_non_std_msg_async(self, *, cls: type[Exception]) -> None:
        async def func(*, value: bool) -> bool:
            if not value:
                raise cls(value)
            await sleep(0.01)
            return not value

        with raises(cls, match="False"):
            _ = await try_await(func(value=False))


if __name__ == "__main__":
    _ = run(
        stream_command('echo "stdout message" && sleep 2 && echo "stderr message" >&2')
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from dacite import Config, from_dict
from hypothesis import given
from hypothesis.strategies import integers, none, sampled_from

from utilities.dacite import yield_literal_forward_references

TruthLit = Literal["true", "false"]  # in 3.12, use type TruthLit = ...


class TestYieldLiteralForwardReferences:
    @given(truth=sampled_from(["true", "false"]))
    def test_literal(self, *, truth: Literal["true", "false"]) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            truth: Literal["true", "false"]

        data = {"truth": truth}
        fwd_refs = dict(yield_literal_forward_references(Example, globalns=globals()))
        result = from_dict(Example, data, config=Config(forward_references=fwd_refs))
        expected = Example(truth=truth)
        assert result == expected

    @given(truth=sampled_from(["true", "false"]) | none())
    def test_literal_nullable(self, *, truth: Literal["true", "false"] | None) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            truth: Literal["true", "false"] | None = None

        data = {"truth": truth}
        fwd_refs = dict(yield_literal_forward_references(Example, globalns=globals()))
        result = from_dict(Example, data, config=Config(forward_references=fwd_refs))
        expected = Example(truth=truth)
        assert result == expected

    @given(truth=sampled_from(["true", "false"]))
    def test_literal_type(self, *, truth: TruthLit) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            truth: TruthLit

        data = {"truth": truth}
        fwd_refs = dict(yield_literal_forward_references(Example, globalns=globals()))
        result = from_dict(Example, data, config=Config(forward_references=fwd_refs))
        expected = Example(truth=truth)
        assert result == expected

    @given(truth=sampled_from(["true", "false"]) | none())
    def test_literal_type_nullable(self, *, truth: TruthLit | None) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            truth: TruthLit | None = None

        data = {"truth": truth}
        fwd_refs = dict(yield_literal_forward_references(Example, globalns=globals()))
        result = from_dict(Example, data, config=Config(forward_references=fwd_refs))
        expected = Example(truth=truth)
        assert result == expected

    @given(n=integers())
    def test_non_literal_type_nullable(self, *, n: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            n: int = 0

        data = {"n": n}
        fwd_refs = dict(yield_literal_forward_references(Example, globalns=globals()))
        result = from_dict(Example, data, config=Config(forward_references=fwd_refs))
        expected = Example(n=n)
        assert result == expected



from __future__ import annotations

from asyncio import iscoroutinefunction
from collections.abc import Callable
from functools import partial, wraps
from typing import Any, TypeVar, cast, overload

from beartype import beartype

_F = TypeVar("_F", bound=Callable[..., Any])


@overload
def beartype_cond(
    func: _F,
    /,
    *,
    setup: Callable[[], bool] | None = ...,
    runtime: Callable[[], bool] | None = ...,
) -> _F: ...
@overload
def beartype_cond(
    func: None = None,
    /,
    *,
    setup: Callable[[], bool] | None = ...,
    runtime: Callable[[], bool] | None = ...,
) -> Callable[[_F], _F]: ...
def beartype_cond(
    func: _F | None = None,
    /,
    *,
    setup: Callable[[], bool] | None = None,
    runtime: Callable[[], bool] | None = None,
) -> _F | Callable[[_F], _F]:
    """Apply `beartype` conditionally."""
    if func is None:
        result = partial(beartype_cond, setup=setup, runtime=runtime)
        return cast(Callable[[_F], _F], result)

    if (setup is not None) and not setup():
        return func

    decorated = beartype(func)
    if runtime is None:
        return decorated

    if not iscoroutinefunction(func):

        @wraps(func)
        def beartype_sync(*args: Any, **kwargs: Any) -> Any:
            if runtime():
                return decorated(*args, **kwargs)
            return func(*args, **kwargs)

        return cast(_F, beartype_sync)

    @wraps(func)
    async def beartype_async(*args: Any, **kwargs: Any) -> Any:
        if runtime():
            return await decorated(*args, **kwargs)
        return await func(*args, **kwargs)

    return cast(_F, beartype_async)


__all__ = ["beartype_cond"]



from __future__ import annotations

import sys
from subprocess import PIPE, CalledProcessError, check_output
from sys import exc_info

from pytest import raises

from tests.conftest import SKIPIF_CI
from utilities.sys import (
    VERSION_MAJOR_MINOR,
    LogExceptionPathsError,
    log_exception_paths,
)
from utilities.text import strip_and_dedent


class TestLogExceptionPaths:
    @SKIPIF_CI
    def test_main(self) -> None:
        code = strip_and_dedent("""
            from __future__ import annotations

            import sys
            from itertools import chain
            from logging import getLogger

            from utilities.sys import log_exception_paths
            from utilities.traceback import trace

            sys.excepthook = log_exception_paths


            @trace
            def first(a: int, b: int, /, *args: int, c: int = 0, **kwargs: int) -> int:
                a *= 2
                b *= 2
                args = tuple(2 * arg for arg in args)
                c *= 2
                kwargs = {k: 2 * v for k, v in kwargs.items()}
                return second(a, b, *args, c, **kwargs)


            @trace
            def second(a: int, b: int, /, *args: int, c: int = 0, **kwargs: int) -> int:
                a *= 2
                b *= 2
                args = tuple(2 * arg for arg in args)
                c *= 2
                kwargs = {k: 2 * v for k, v in kwargs.items()}
                result = sum(chain([a, b], args, [c], kwargs.values()))
                assert result % 10 == 0, f"Result ({result}) must be divisible by 10"
                return result


            _ = first(1, 2, 3, 4, c=5, d=6, e=7)
            """)
        with raises(CalledProcessError) as exc_info:
            _ = check_output([sys.executable, "-c", code], stderr=PIPE, text=True)
        stderr = exc_info.value.stderr.strip("\n")
        expected = strip_and_dedent("""
            ExcPath(
                frames=[
                    _Frame(
                        module='__main__',
                        name='first',
                        code_line='',
                        line_num=20,
                        args=(1, 2, 3, 4),
                        kwargs={'c': 5, 'd': 6, 'e': 7},
                        locals={
                            'a': 2,
                            'b': 4,
                            'c': 10,
                            'args': (6, 8),
                            'kwargs': {'d': 12, 'e': 14}
                        }
                    ),
                    _Frame(
                        module='__main__',
                        name='second',
                        code_line='',
                        line_num=31,
                        args=(2, 4, 6, 8, 10),
                        kwargs={'d': 12, 'e': 14},
                        locals={
                            'a': 4,
                            'b': 8,
                            'c': 0,
                            'args': (12, 16, 20),
                            'kwargs': {'d': 24, 'e': 28},
                            'result': 112
                        }
                    )
                ],
                error=AssertionError('Result (112) must be divisible by 10')
            )
            """)
        assert stderr == expected

    def test_non_error(self) -> None:
        exc_type, exc_val, traceback = exc_info()
        with raises(LogExceptionPathsError, match="No exception to log"):
            log_exception_paths(exc_type, exc_val, traceback)


class TestVersionMajorMinor:
    def test_main(self) -> None:
        assert isinstance(VERSION_MAJOR_MINOR, tuple)
        expected = 2
        assert len(VERSION_MAJOR_MINOR) == expected



from __future__ import annotations

from dataclasses import dataclass
from logging import Logger, getLogger
from sys import version_info
from typing import TYPE_CHECKING

from typing_extensions import override

from utilities.traceback import assemble_exception_paths

if TYPE_CHECKING:
    from types import TracebackType

_LOGGER = getLogger(__name__)
VERSION_MAJOR_MINOR = (version_info.major, version_info.minor)


def log_exception_paths(
    exc_type: type[BaseException] | None,
    exc_val: BaseException | None,
    traceback: TracebackType | None,
    /,
    *,
    logger: Logger = _LOGGER,
    max_width: int = 80,
    indent_size: int = 4,
    max_length: int | None = None,
    max_string: int | None = None,
    max_depth: int | None = None,
    expand_all: bool = False,
) -> None:
    """Exception hook to log the traceback."""
    _ = (exc_type, traceback)  # skipif-ci-and-windows
    if exc_val is None:  # pragma: no cover
        raise LogExceptionPathsError
    error = assemble_exception_paths(exc_val)  # skipif-ci-and-windows
    try:  # skipif-ci-and-windows
        from rich.pretty import pretty_repr
    except ImportError:  # pragma: no cover
        repr_use = repr(error)
    else:  # skipif-ci-and-windows
        repr_use = pretty_repr(
            error,
            max_width=max_width,
            indent_size=indent_size,
            max_length=max_length,
            max_string=max_string,
            max_depth=max_depth,
            expand_all=expand_all,
        )
    logger.error("%s", repr_use)  # skipif-ci-and-windows


@dataclass(kw_only=True, slots=True)
class LogExceptionPathsError(Exception):
    @override
    def __str__(self) -> str:
        return "No exception to log"


__all__ = ["VERSION_MAJOR_MINOR", "LogExceptionPathsError", "log_exception_paths"]
